# Copyright (c) 2016-2019, TDK Electronics
# All rights reserved.
#
# Author: Marek Cieplucha, https://github.com/mciepluc
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met 
# (The BSD 2-Clause License):
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation 
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from setuptools import setup
from setuptools import find_packages

import cocotb_coverage
version = cocotb_coverage.__version__

from os import path

def read_file(fname):
    return open(path.join(path.dirname(__file__), fname)).read()

setup(
    name='cocotb-coverage',
    version=version,
    description='Functional Coverage and Constrained Randomization Extensions for Cocotb',
    url='https://github.com/mciepluc/cocotb-coverage',
    license='BSD',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    author='Marek Cieplucha',
    author_email='',
    packages=find_packages(),
    install_requires= ['cocotb', 'python-constraint'],
    python_requires=">=3.3",
    platforms='any',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
    ],
)



# Copyright (c) 2016-2019, TDK Electronics
# All rights reserved.
#
# Author: Marek Cieplucha, https://github.com/mciepluc
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met 
# (The BSD 2-Clause License):
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation 
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Constrained-random verification features.

Classes:

* :class:`Randomized` - base class for randoimzed types.

"""

import random
import inspect
import itertools
import warnings

# python-constraint is an external pip-installable package used here
import constraint

class Randomized(object):
    """Base class for randomized types.

    The final class should contain defined random variables using the 
    :meth:`add_rand()` method.

    Constraints may be added and deleted using the
    :meth:`add_constraint()` and :meth:`del_constraint()` methods respectively.

    A constraint is an arbitrary function and may either return a 
    ``True``/``False`` value (*hard constraints*) or a numeric value, which may
    be interpreted as *soft constraints* or *distribution functions*.

    Constraint function arguments (names) must match final class attributes 
    (random or not). Constraints may have multiple random arguments which 
    corresponds to multi-dimensional distributions.

    The function :meth:`randomize()` performs a randomization for all random 
    variables meeting all defined constraints.

    The function :meth:`randomize_with()` performs a randomization using 
    additional constraint functions given in an argument.

    The functions :meth:`pre_randomize()` and :meth:`post_randomize()` are 
    called before and after :meth:`randomize` and should be overloaded in a 
    final class if necessary.

    If hard constraint cannot be resolved, an exception is thrown. If a soft
    constraint cannot be resolved (all acceptable solutions have zero 
    probability), then the variable value is not being randomized.

    Example:

    >>> class FinalRandomized(Randomized):
    >>>   def __init__(self, x):
    >>>       Randomized.__init__(self)
    >>>       self.x = x
    >>>       self.y = 0
    >>>       self.z = 0
    >>>
    >>>       # define y as a random variable taking values from 0 to 9
    >>>       add_rand("y", list(range(10)))
    >>>
    >>>       # define z as a random variable taking values from 0 to 4
    >>>       add_rand("z", list(range(5)))
    >>>
    >>>       # hard constraint
    >>>       add_constraint(lambda x, y: x !=y) 
    >>>       # multi-dimensional distribution
    >>>       add_constraint(lambda y, z: y + z) 
    >>>
    >>> # create randomized object instance (default values at this point)
    >>> obj_ = FinalRandomized(5)
    >>> # randomize object with additional contraint 
    >>> obj_.randomize_with(lambda z : z > 3)  

    As generating constrained random objects may involve a lot of computations,
    it is recommended to limit random variables domains and use
    :meth:`pre_randomize()`/:meth:`post_randomize()` methods where possible.
    """

    def __init__(self):
        # all random variables, map NAME -> DOMAIN
        self._randVariables = {}

        # all simple constraints: functions of single random variable and
        # optional non-random variables
        # map VARIABLE NAME -> FUNCTION
        self._simpleConstraints = {}

        # all implicit constraints: functions that requires to be resolved by a
        # Solver
        # map TUPLE OF VARIABLE NAMES -> FUNCTION
        self._implConstraints = {}

        # all implicit distributions: functions that involve implicit random
        # variables and single unconstrained variable
        # map TUPLE OF VARIABLE NAMES -> FUNCTION
        self._implDistributions = {}

        # all simple distributions: functions of unconstrained random variables
        # and non-random variables
        # map VARIABLE NAME -> FUNCTION
        self._simpleDistributions = {}

        # list of lists containing random variables solving order
        self._solve_order = []

    def add_rand(self, var, domain=None):
        """Add a random variable to the solver.

        All random variables must be defined before adding any constraint with 
        :meth:`add_constraint`. Therefore it is highly recommended to call 
        ``add_rand`` in the ``__init__`` method of your final class.

        Args:
            var (str): a variable name corresponding to the class member 
                variable.
            domain (list, optional): a list of all allowed values of the 
                variable ``var``. By default, a list with values ``0`` to 
                ``65534`` (16 bit unsigned int domain) is used.

        Examples:

        >>> add_rand("data", list(range(1024)))
        >>> add_rand("delay", ["small", "medium", "high"])
        """
        assert (not (self._simpleConstraints or
                     self._implConstraints or
                     self._implDistributions or
                     self._simpleDistributions)
                ), \
            "All random variables must be defined before adding a constraint."

        if not domain:
            domain = range(65535)  # 16 bit unsigned int

        self._randVariables[var] = domain  # add a variable to the map

    def add_constraint(self, cstr):
        """Add a constraint function to the solver.

        A constraint may return ``True``/``False`` or a numeric value.
        Constraint function arguments must be valid class member names (random 
        or not). Arguments must be listed in alphabetical order.

        Due to calculation complexity, it is recommended to create as few 
        constraints as possible and implement
        :meth:`pre_randomize()`/:meth:`post_randomize()` methods, or use the 
        :meth:`solve_order()` function.

        Each constraint is associated with its arguments being random 
        variables,which means for each random variable combination only one 
        constraint of the ``True``/``False`` type and one numeric may be 
        defined. The latter will overwrite the existing one.

        For example, when class has two random variables ``(x, y)``, 
        six constraint functions may be defined: boolean and numeric 
        constraints of ``x``, ``y`` and a pair ``(x, y)``.

        Args:
            cstr (func): a constraint function.

        Returns:
            func or None: an overwritten constraint or ``None`` if no 
            overwrite happened.

        Examples:

        >>> def highdelay_cstr(delay):
        >>>     return delay == "high"
        >>>
        >>> add_constraint(highdelay_cstr)  # hard constraint
        >>> add_constraint(lambda data : data < 128)  # hard constraint
        >>>
        >>> # distribution (highest probability density at the boundaries):
        >>> add_constraint(lambda data : abs(64 - data))
        >>>
        >>> # hard constraint of multiple variables (some of them may be 
        >>> # non-random):
        >>> add_constraint(lambda x,y,z : x + y + z == 0)
        >>>
        >>> # soft constraint created by applying low probability density for 
        >>> # some solutions:
        >>> add_constraint(
        >>>  lambda delay, size : 0.01 if (size < 5 & delay == "medium") else 1
        >>> )
        >>> # constraint that overwrites the previously defined one
        >>> # (data < 128)
        >>> add_constraint(lambda data : data < 256)
        """

        # just add constraint considering all random variables
        return self._add_constraint(cstr, self._randVariables)

    def solve_order(self, *orderedVars):
        """Define an order of the constraints resolving.

        Constraints are being resolved in a given order, which means that 
        randomization is called in separated steps, where at each next step
        some constraints are already resolved. Number of arguments defines 
        number of the randomization steps.

        Args:
            *orderedVars (multiple str or list): Variables that are requested 
                to be resolved in an specific order.

        Example:

        >>> add_rand("x", list(range(0,10)))
        >>> add_rand("y", list(range(0,10)))
        >>> add_rand("z", list(range(0,10)))
        >>> add_rand("w", list(range(0,10)))
        >>> add_constraint(lambda x, y : x + y = 9)
        >>> add_constraint(lambda z : z < 5)
        >>> add_constraint(lambda w : w > 5)
        >>>
        >>> solve_order(["x", "z"], "y")
        >>> # In a first step, "z", "x" and "w" will be resolved, which means 
        >>> # only the second and third constraint will be applied. In a second 
        >>> # step, the first constraint will be resolved as it was requested 
        >>> # to solve "y" after "x" and "z". "x" will be interpreted as a 
        >>> # constant in this case.
        """
        self._solve_order = []
        for selRVars in orderedVars:
            if type(selRVars) is not list:
                self._solve_order.append([selRVars])
            else:
                self._solve_order.append(selRVars)

    def del_constraint(self, cstr):
        """Delete a constraint function.

        Args:
            cstr (func): a constraint function.

        Example:

        >>> del_constraint(highdelay_cstr)
        """
        return self._del_constraint(cstr, self._randVariables)

    def pre_randomize(self):
        """A function that is called before 
        :meth:`randomize`/:meth:`randomize_with`.

        To be overridden in a final class if used.
        """
        pass

    def post_randomize(self):
        """A function that is called after 
        :meth:`randomize`/:meth:`randomize_with`.

        To be overridden in a final class if used.
        """
        pass

    def randomize(self):
        """Randomize a final class using only predefined constraints."""
        self._randomize()

    def randomize_with(self, *constraints):
        """Randomize a final class using the additional constraints given.

        Additional constraints may override existing ones.

        Args:
            *constraints ((multiple) func): additional constraints to be 
                applied.

        """
        overwritten_constrains = []

        # add new constraints
        for cstr in constraints:
            overwritten = self.add_constraint(cstr)
            if overwritten:
                overwritten_constrains.append(overwritten)

        raise_exception = False
        try:
            self._randomize()
        except:
            raise_exception = True

        # remove new constraints
        for cstr in constraints:
            self.del_constraint(cstr)

        # add back overwritten constraints
        for cstr in overwritten_constrains:
            self.add_constraint(cstr)

        if raise_exception:
            raise Exception("Could not resolve implicit constraints!")

    def _add_constraint(self, cstr, rvars):
        """Add a constraint for a specific random variables list
        (which determines a type of a constraint - simple or implicit).
        """
        if isinstance(cstr, constraint.Constraint):
            # could be a Constraint object...
            pass
        else:
            variables = inspect.signature(cstr).parameters
            assert (list(variables) == sorted(variables)), \
                "Variables of a constraint function must be defined in \
                alphabetical order"

            # determine the function type... rather unpythonic but necessary 
            # for distinction between a constraint and a distribution
            callargs = []
            rand_variables = []
            for var in variables:
                if var in rvars:
                    rand_variables.append(var)
                    callargs.append(random.choice(rvars[var]))
                else:
                    callargs.append(getattr(self, var))

            ret = cstr(*callargs)

            def _addToMap(_key, _map):
                overwriting = None
                if _key in _map:
                    overwriting = _map[_key]
                _map[_key] = cstr
                return overwriting

            if type(ret) is bool:
                # this is a constraint
                if (len(rand_variables) == 1):
                    overwriting = _addToMap(
                        rand_variables[0], self._simpleConstraints)
                else:
                    overwriting = _addToMap(
                        tuple(rand_variables), self._implConstraints)
            else:
                # this is a distribution
                if (len(rand_variables) == 1):
                    overwriting = _addToMap(
                        rand_variables[0], self._simpleDistributions)
                else:
                    overwriting = _addToMap(
                        tuple(rand_variables), self._implDistributions)

            return overwriting

    def _del_constraint(self, cstr, rvars):
        """Delete a constraint for a specific random variables list
        (which determines a type of a constraint - simple or implicit).
        """
        if isinstance(cstr, constraint.Constraint):
            # could be a Constraint object...
            pass
        else:
            variables = inspect.signature(cstr).parameters

            rand_variables = [
                var for var in variables if var in rvars]

            if (len(rand_variables) == 1):
                if rand_variables[0] in self._simpleConstraints:
                    del self._simpleConstraints[rand_variables[0]]
                elif rand_variables[0] in self._simpleDistributions:
                    del self._simpleDistributions[rand_variables[0]]
                else:
                    assert(0), "Could not delete a constraint!"
            else:
                if tuple(rand_variables) in self._implConstraints:
                    del self._implConstraints[tuple(rand_variables)]
                elif tuple(rand_variables) in self._implDistributions:
                    del self._implDistributions[tuple(rand_variables)]
                else:
                    assert(0), "Could not delete a constraint!"


    def _randomize(self):
        """Call :meth:`_resolve` and 
        :meth:`pre_randomize`/:meth:`post_randomize` functions with respect to 
        defined variables resolving order.
        """

        self.pre_randomize()
        if not self._solve_order:
            #call _resolve for all random variables
            solution = self._resolve(self._randVariables)
            self._update_variables(solution)
        else:

            #list of random variables names
            remainingRVars = list(self._randVariables.keys())

            #list of resolved random variables names
            resolvedRVars = []

            #list of random variables with defined solve order
            remainingOrderedRVars = [item for sublist in self._solve_order
                                     for item in sublist]

            allConstraints = [] # list of functions (all constraints and dstr)
            allConstraints.extend([self._implConstraints[_]
                               for _ in self._implConstraints])
            allConstraints.extend([self._implDistributions[_]
                               for _ in self._implDistributions])
            allConstraints.extend([self._simpleConstraints[_]
                               for _ in self._simpleConstraints])
            allConstraints.extend([self._simpleDistributions[_]
                               for _ in self._simpleDistributions])

            for selRVars in self._solve_order:

                #step 1: determine all variables to be solved at this stage
                actualRVars = list(selRVars) #add selected
                for rvar in actualRVars:
                    remainingOrderedRVars.remove(rvar) #remove selected
                    remainingRVars.remove(rvar) #remove selected

                #if implicit constraint requires a variable which is not given
                #at this stage, it will be resolved later
                for rvar in remainingRVars:
                    rvar_unused = True
                    for c_vars in self._implConstraints:
                        if rvar in c_vars:
                            rvar_unused = False
                    for d_vars in self._implDistributions:
                        if rvar in d_vars:
                            rvar_unused = False
                    if rvar_unused and not rvar in remainingOrderedRVars:
                        actualRVars.append(rvar)
                        remainingRVars.remove(rvar)

                # a new map of random variables
                newRandVariables = {}
                for var in self._randVariables:
                    if var in actualRVars:
                        newRandVariables[var] = self._randVariables[var]

                #step 2: select only valid constraints at this stage

                #delete all constraints and add back but considering only
                #limited list of random vars
                actualCstr = []
                for f_cstr in allConstraints:
                    self.del_constraint(f_cstr)
                    f_cstr_args = inspect.signature(f_cstr).parameters
                    #add only constraints containing actualRVars but not
                    #remainingRVars
                    add_cstr = True
                    for var in f_cstr_args:
                        if (var in self._randVariables and
                            not var in resolvedRVars and
                            (not var in actualRVars or var in remainingRVars)
                            ):
                            add_cstr = False
                    if add_cstr:
                        self._add_constraint(f_cstr, newRandVariables)
                        actualCstr.append(f_cstr)

                #call _resolve for all random variables
                solution = self._resolve(newRandVariables)
                self._update_variables(solution)

                resolvedRVars.extend(actualRVars)

                #add back everything as it was before this stage
                for f_cstr in actualCstr:
                    self._del_constraint(f_cstr, newRandVariables)

                for f_cstr in allConstraints:
                    self._add_constraint(f_cstr, self._randVariables)

        self.post_randomize()

    def _resolve(self, randomVariables):
        """Resolve constraints for given random variables."""

        # we need a copy, as we will be updating domains
        randVariables = dict(randomVariables)

        # step 1: determine search space by applying simple constraints to the
        # random variables

        for rvar in randVariables:
            domain = randVariables[rvar]
            new_domain = []
            if rvar in self._simpleConstraints:
                # a simple constraint function to be applied
                f_cstr = self._simpleConstraints[rvar]
                # check if we have non-random vars in cstr...
                # arguments of the constraint function
                f_c_args = inspect.signature(f_cstr).parameters
                for ii in domain:
                    f_cstr_callvals = []
                    for f_c_arg in f_c_args:
                        if (f_c_arg == rvar):
                            f_cstr_callvals.append(ii)
                        else:
                            f_cstr_callvals.append(getattr(self, f_c_arg))
                    # call simple constraint for each domain element
                    if f_cstr(*f_cstr_callvals):
                        new_domain.append(ii)
                # update the domain with the constrained one
                randVariables[rvar] = new_domain

        # step 2: resolve implicit constraints using external solver

        # external hard constraint solver - package python-constraint
        problem = constraint.Problem()

        constrainedVars = []  # all random variables for the solver

        for rvars in self._implConstraints:
            # add all random variables
            for rvar in rvars:
                if not rvar in constrainedVars:
                    problem.addVariable(rvar, randVariables[rvar])
                    constrainedVars.append(rvar)
            # add constraint
            problem.addConstraint(self._implConstraints[rvars], rvars)

        # solve problem
        solutions = problem.getSolutions()

        if (len(solutions) == 0) & (len(constrainedVars) > 0):
            raise Exception("Could not resolve implicit constraints!")

        # step 3: calculate implicit distributions for all random variables
        # except simple distributions

        # all variables that have defined distribution functions
        distrVars = []
        # solutions with applied distribution weights - list of maps VARIABLE
        # -> VALUE
        dsolutions = []

        # add all variables that have defined distribution functions
        for dvars in self._implDistributions:
            # add all variables that have defined distribution functions
            for dvar in dvars:
                if dvar not in distrVars:
                    distrVars.append(dvar)

        # all variables that have defined distributions but unconstrained
        ducVars = [var for var in distrVars if var not in constrainedVars]

        # list of domains of random unconstrained variables
        ducDomains = [randVariables[var] for var in ducVars]

        # Cartesian product of above
        ducSolutions = list(itertools.product(*ducDomains))

        # merge solutions: constrained ones and all possible distribution 
        # values
        for sol in solutions:
            for ducsol in ducSolutions:
                dsol = dict(sol)
                jj = 0
                for var in ducVars:
                    dsol[var] = ducsol[jj]
                    jj += 1
                dsolutions.append(dsol)

        dsolution_weights = []
        dsolutions_reduced = []

        for dsol in dsolutions:  # take each solution
            weight = 1.0
            # for all defined implicit distributions
            for dstr in self._implDistributions:
                f_idstr = self._implDistributions[dstr]
                f_id_args = inspect.signature(f_idstr).parameters
                # all variables in solution we need to calculate weight
                f_id_callvals = []
                for f_id_arg in f_id_args:  # for each variable name
                    if f_id_arg in dsol:  # if exists in solution
                        f_id_callvals.append(dsol[f_id_arg])
                    else:  # get as non-random variable
                        f_id_callvals.append(getattr(self, f_id_arg))
                # update weight of the solution - call distribution function
                weight = weight * f_idstr(*f_id_callvals)
            # do the same for simple distributions
            for dstr in self._simpleDistributions:
                # but only if variable is already in the solution
                # if it is not, it will be calculated in step 4
                if dstr in sol:
                    f_sdstr = self._simpleDistributions[dstr]
                    f_sd_args = inspect.signature(f_sdstr).parameters
                    # all variables in solution we need to calculate weight
                    f_sd_callvals = []
                    for f_sd_arg in f_sd_args:  # for each variable name
                        if f_sd_arg in dsol:  # if exists in solution
                            f_sd_callvals.append(dsol[f_sd_arg])
                        else:  # get as non-random variable
                            f_sd_callvals.append(getattr(self, f_sd_arg))
                    # update weight of the solution - call distribution 
                    # function
                    weight = weight * f_sdstr(*f_sd_callvals)
            if (weight > 0.0):
                dsolution_weights.append(weight)
                # remove solutions with weight = 0
                dsolutions_reduced.append(dsol)

        solution_choice = self._weighted_choice(
            dsolutions_reduced, dsolution_weights)
        solution = solution_choice if solution_choice is not None else {}

        # step 4: calculate simple distributions for remaining random variables
        for dvar in randVariables:
            if not dvar in solution:  # must be already unresolved variable
                domain = randVariables[dvar]
                weights = []
                if dvar in self._simpleDistributions:
                    # a simple distribution to be applied
                    f_dstr = self._simpleDistributions[dvar]
                    # check if we have non-random vars in dstr...
                    f_d_args = inspect.signature(f_dstr).parameters
                    # list of lists of values for function call
                    f_d_callvals = []
                    for i in domain:
                        f_d_callval = []
                        for f_d_arg in f_d_args:
                            if (f_d_arg == dvar):
                                f_d_callval.append(i)
                            else:
                                f_d_callval.append(getattr(self, f_d_arg))
                        f_d_callvals.append(f_d_callval)
                    # call distribution function for each domain element to get
                    # the weight
                    weights = [f_dstr(*f_d_callvals_i)
                               for f_d_callvals_i in f_d_callvals]
                    new_solution = self._weighted_choice(domain, weights)
                    if new_solution is not None:
                        # append chosen value to the solution
                        solution[dvar] = new_solution
                else:
                    # random variable has no defined distribution function -
                    # call simple random.choice
                    solution[dvar] = random.choice(domain)

        return solution

    def _weighted_choice(self, solutions, weights):
        """Get a solution from the list with defined weights."""
        try:
            import numpy
            # pick weighted random
            return numpy.random.choice(solutions, size=1, p=weights)
        except:
            # if numpy not available
            non_zero_weights = [x for x in weights if x > 0]

            if not non_zero_weights:
                return None

            min_weight = min(non_zero_weights)

            weighted_solutions = []

            for x in range(len(solutions)):
                # insert each solution to the list multiple times
                weighted_solutions.extend(
                    [solutions[x] for _ in range(
                        int(weights[x] * (1.0 / min_weight)))
                     ])

            return random.choice(weighted_solutions)

    def _update_variables(self, solution):
        """Update members of the final class after randomization."""
        # update class members
        for var in self._randVariables:
            if var in solution:
                setattr(self, var, solution[var])

    #deprecated
    def addRand(self, var, domain=None):
        """.. deprecated:: 1.0"""
        warnings.warn(
         "Function addRand() is deprecated, use add_rand() instead"
        )
        self.add_rand(var, domain)

    def solveOrder(self, *orderedVars):
        """.. deprecated:: 1.0"""
        warnings.warn(
         "Function solveOrder() is deprecated, use solve_order() instead"
        )
        self.solve_order(*orderedVars)

    def addConstraint(self, cstr):
        """.. deprecated:: 1.0"""
        warnings.warn(
         "Function addConstraint() is deprecated, use add_constraint() instead"
        )
        self.add_constraint(cstr)

    def delConstraint(self, cstr):
        """.. deprecated:: 1.0"""
        warnings.warn(
         "Function delConstraint() is deprecated, use del_constraint() instead"
        )
        self.del_constraint(cstr)





# Copyright (c) 2016-2019, TDK Electronics
# All rights reserved.
#
# Author: Marek Cieplucha, https://github.com/mciepluc
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met
# (The BSD 2-Clause License):
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Functional Coverage features.

Classes:

* :class:`CoverageDB` - singleton containing coverage database.
* :class:`CoverItem` - base class for coverage, corresponds to a covergroup,
  created automatically.
* :class:`CoverPoint` - a cover point with bins.
* :class:`CoverCross` - a cover cross of cover points.
* :class:`CoverCheck` - a cover point which checks only a pass/fail condition.

Functions:

* :func:`~.coverage_section` - allows for convenient definition of multiple
  coverage items and combines them into a single decorator.
* :func:`~.XML_merger` - merges coverage databases saved in XML format. 
"""

from functools import wraps
from collections import OrderedDict
import inspect
import operator
import itertools
import warnings
import copy
from xml.etree import ElementTree as et


class CoverageDB(dict):
    """ Class (singleton) containing coverage database.

    This is the coverage prefix tree (trie) containing all coverage objects
    with name string as a key (using dot as a stage separator). Coverage
    primitives may be accessed by string identificator. Example coverage trie
    is shown below:

    .. image:: coverstruct.png
        :scale: 60 %
        :align: center

    Examples:

    >>> CoverageDB()["top"] #access whole coverage under ``top``
    >>> CoverageDB()["top.b"] #access whole covergroup
    >>> CoverageDB()["top.b.cp1"] #access specific coverpoint
    """
    _instance = None

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = dict.__new__(class_, *args, **kwargs)
        return class_._instance

    def report_coverage(self, logger, bins=False, node=""):
        """Print sorted coverage with optional bins details.

        Args:
            logger (func): a logger object.
            bins (bool, optional): print bins details.
            node (str, optional): starting node of the coverage trie.
        """
        sorted_cov = sorted(self, key=str.lower)
        for ii in filter(lambda _ : _.startswith(node), sorted_cov):
            logger("   " * ii.count('.') + "%s : %s, coverage=%d, size=%d " %
                   (ii, self[ii], self[ii].coverage, self[ii].size)
                   )
            if (type(self[ii]) is not CoverItem) & (bins):
                for jj in self[ii].detailed_coverage:
                    logger("   " * ii.count('.') + "   BIN %s : %s" %
                           (jj, self[ii].detailed_coverage[jj])
                           )

    def export_to_xml(self, xml_name='coverage.xml'):
        """Export coverage_db to xml document.

        Args:
            xml_name (str): output document name with .xml suffix
        """
        xml_db_dict = {}

        def create_top():
            attrib_dict = {'abs_name': 'top'}
            if 'top' in self:
                attrib_dict['size'] = str(self['top'].size)
                attrib_dict['coverage'] = str(self['top'].coverage)
                attrib_dict['cover_percentage'] = str(round(
                    self['top'].cover_percentage, 2))
            xml_db_dict['top'] = et.Element('top', attrib=attrib_dict)

        def create_element(name_elem, parent, name_elem_full):
            attrib_dict = {}
            prefix = '' if 'top' in self else 'top.'

            # Common attributes
            attrib_dict['size'] = str(self[name_elem_full].size)
            attrib_dict['coverage'] = str(self[name_elem_full].coverage)
            attrib_dict['cover_percentage'] = str(round(
                self[name_elem_full].cover_percentage, 2))
            attrib_dict['abs_name'] = prefix+name_elem_full
            if (type(self[name_elem_full]) is not CoverItem):
                attrib_dict['weight'] = str(self[name_elem_full].weight)
                attrib_dict['at_least'] = str(
                    self[name_elem_full].at_least)

            # Create element: xml_db_dict[a.b.c] = et(a.b (parent), c(element))
            xml_db_dict[name_elem_full] = et.SubElement(xml_db_dict[parent],
                                                        name_elem,
                                                        attrib=attrib_dict)

            # Create bins for CoverCross and CoverPoint
            if type(self[name_elem_full]) is not CoverItem:
                bin_count = 0
                # Database in format: key == bin, value == no_of_hits
                for key, value in self[name_elem_full].detailed_coverage.items():
                    attrib_dict.clear()
                    attrib_dict['bin'] = str(key)
                    attrib_dict['hits'] = str(value)
                    attrib_dict['abs_name'] = (prefix+name_elem_full
                                               +'.bin'+str(bin_count))
                    xml_db_dict[name_elem_full+'.bin'+str(bin_count)] = (
                        et.SubElement(xml_db_dict[name_elem_full],
                                      'bin'+str(bin_count),
                                      attrib=attrib_dict))
                    bin_count += 1

        # ======================== Function body ==============================
        create_top()
        for name_elem_full in self:
            if name_elem_full == 'top':
                pass
            else:
                name_list = name_elem_full.split('.')
                name_elem = name_list[-1]
                name_parent = '.'.join(name_list[:-1])
                if name_parent == '':
                    name_parent = 'top'
                create_element(name_elem, name_parent, name_elem_full)

        # update total coverage if there was no 'top' in coverage_db
        if xml_db_dict['top'].attrib == {'abs_name': 'top'}:
            top_size = 0
            top_coverage = 0
            for child in xml_db_dict['top']:
                top_size += int(child.attrib['size'])
                top_coverage += int(child.attrib['coverage'])
                top_cover_percentage = round(top_coverage*100/top_size, 2)
            xml_db_dict['top'].set('size', str(top_size))
            xml_db_dict['top'].set('coverage', str(top_coverage))
            xml_db_dict['top'].set(
                'cover_percentage', str(top_cover_percentage))

        root = et.ElementTree(xml_db_dict['top']).getroot()
        _indent(root)
        et.ElementTree(xml_db_dict['top']).write(xml_name)

# global variable collecting coverage in a prefix tree (trie)
coverage_db = CoverageDB()
""" Instance of the :class:`CoverageDB`."""

class CoverItem(object):
    """Class used to describe coverage groups.

    ``CoverItem`` objects are created automatically. This is a base class for
    all coverage primitives (:class:`CoverPoint`, :class:`CoverCross` or
    :class:`CoverCheck`). It may be used as a base class for other,
    user-defined coverage types. 
    """

    def __init__(self, name):
        self._name = name
        self._size = 0
        self._coverage = 0
        self._parent = None
        self._children = []
        self._new_hits = []
        self._weight = 0
        self._at_least = 0

        self._threshold_callbacks = {}
        self._bins_callbacks = {}

        # check if parent exists
        if "." in name:
            parent_name = ".".join(name.split(".")[:-1])
            if not parent_name in coverage_db:
                CoverItem(name=parent_name)

            self._parent = coverage_db[parent_name]
            self._parent._children.append(self)

        coverage_db[name] = self

    def _update_coverage(self, coverage):
        """Update the parent coverage level as requested by derived classes. 
        """
        current_coverage = self._coverage
        self._coverage += coverage
        if self._parent is not None:
            self._parent._update_coverage(coverage)

        # notify callbacks
        for ii in self._threshold_callbacks:
            if (ii > 100 * current_coverage / self.size
                    and ii <= self.cover_percentage):
                self._threshold_callbacks[ii]()

    def _update_size(self, size):
        """Update the parent size as requested by derived classes. 
        """
        self._size += size
        if self._parent is not None:
            self._parent._update_size(size)

    def add_threshold_callback(self, callback, threshold):
        """Add a threshold callback to the :class:`CoverItem` or any its 
        derived class. 

        A callback is called (once) when the threshold is crossed, so that 
        coverage level of this particular cover group (or other object) exceeds 
        defined % value. 

        Args:
            callback (func): a callback function.
            threshold (int): a callback call threshold (% coverage). 

        Examples:

        >>> def notify_threshold():
        >>>     print("reached 50% coverage!")
        >>>
        >>> # add callback to the cover group
        >>> coverage_db["top.covergroup1"].add_threshold_callback(
        >>>   notify_threshold, 50
        >>> )
        >>> # add callback to the cover point
        >>> coverage_db["top.covergroup1.coverpoint4"].add_threshold_callback(
        >>>   notify_threshold, 50
        >>> )
        """
        self._threshold_callbacks[threshold] = callback

    def add_bins_callback(self, callback, bins):
        """Add a bins callback to the derived class of the :class:`CoverItem`.

        A callback is called (once) when a specific bin is covered. 

        Args:
            callback (func): a callback function.
            bins: a particular bin (type depends on bins type). 

        Examples:

        >>> def notify_bins():
        >>>     print("covered bin 'special case'")
        >>>
        >>> coverage_db["top.covergroup1.coverpoint5"].add_bins_callback(
        >>>   notify_bins, 'special case'
        >>> )
        """
        self._bins_callbacks[bins] = callback

    @property
    def size(self):
        """Return size of the coverage primitive.

        Size of the cover group (or other coverage primitive) is returned. This
        is a total number of bins associated with assigned weights. 

        Returns:
            int: size of the coverage primitive.
        """
        return self._size

    @property
    def coverage(self):
        """Return size of the covered bins in the coverage primitive.

        Number of the covered bins in cover group (or other coverage primitive) 
        is returned. This is a number of covered bins associated with assigned 
        weights. 

        Returns:
            int: size of the covered bins.
        """
        return self._coverage

    @property
    def cover_percentage(self):
        """Return coverage level of the coverage primitive.

        Percent of the covered bins in cover group (or other coverage 
        primitive) is returned. This is basically a :meth:`coverage()` divided
        by :meth:`size()` in %.

        Returns:
            float: percent of the coverage.
        """
        return 100 * self.coverage / self.size

    @property
    def detailed_coverage(self):
        """Return detailed coverage - full list of bins associated with number
        of hits. If labels are assigned to bins, labels are returned instead
        of bins values.

        A dictionary (bins) -> (number of hits) is returned. 

        Returns:
            dict: dictionary associating number of hits with a particular bins.
        """
        coverage = {}
        for child in self._children:
            coverage.append(child.detailed_coverage)
        return coverage

    @property
    def new_hits(self):
        """Return bins hit at last sampling event. Works only for objects 
        deriving from :class:`CoverItem`.

        Returns:
            list: list of the new bins (which have not been already covered)
            sampled at last sampling event.
        """
        return self._new_hits

    @property
    def weight(self):
        """Return weight of the coverage primitive. Works only for objects 
        deriving from :class:`CoverItem`.

        Returns:
            int: weight of the coverage primitive.
        """
        return self._weight

    @property
    def at_least(self):
        """Return ``at_least`` attribute of the coverage primitive. Works only 
        for objects deriving from :class:`CoverItem`.

        Returns:
            int: ``at_least`` attribute of the coverage primitive.
        """
        return self._at_least


class CoverPoint(CoverItem):
    """Class used to create coverage points as decorators. 

    This decorator samples members of the decorated function (its signature). 
    Sampling matches predefined bins according to the rule:
    ``rel(xf(args), bin) == True``

    Args:
        name (str): a ``CoverPoint`` path and name, defining its position in a 
            coverage trie.
        vname (str, optional): a name of the variable to be covered (use this 
            only when covering a *single* variable in the decorated function
            signature).
        xf (func, optional): a transformation function which transforms 
            arguments of the decorated function. If ``vname`` and ``xf`` are 
            not defined, matched is a single input argument (if only one 
            exists) or a tuple (if multiple exist). Note that the ``self`` 
            argument is *always* removed from the argument list. 
        bins (list): a list of bins objects to be matched. Note that for 
            non-trivial types, a ``rel`` must always be defined (or the 
            equality operator must be overloaded).
        bins_labels (list, optional): a list of labels (str) associated with
            defined bins. Both lists lengths must match.
        rel (func, optional): a relation function which defines the bins 
            matching relation (by default, the equality operator ``==``).
        weight (int, optional): a ``CoverPoint`` weight (by default ``1``).
        at_least (int, optional): the number of hits per bins to be considered 
            as covered (by default ``1``).
        inj (bool, optional): "injection" feature, defines that more than a 
            single bin can be matched at one sampling (default ``False``).

    Example:

    >>> @coverage.CoverPoint( # cover (arg/2) < 1...5 (5 bins)
    ...     name = "top.parent.coverpoint1", 
    ...     xf = lambda x : x/2, 
    ...     rel = lambda x, y : x < y, 
    ...     bins = list(range(1, 5))
    ... )
    >>> @coverage.CoverPoint( # cover (arg) == 1...5 (5 bins)
    ...     name = "top.parent.coverpoint2", 
    ...     vname = "arg",
    ...     bins = list(range(1, 5))
    ... )
    >>> def decorated_func1(self, arg):
    ...     ...

    >>> @coverage.CoverPoint( # cover (arg1, arg2) == (1, 1) or (0, 0) (2 bins)
    ...     name = "top.parent.coverpoint3", 
    ...     bins = [(1, 1), (0, 0)]
    ... )
    >>> def decorated_func1(self, arg1, arg2):
    ...     ...
    """

    # conditional Object creation, only if name not already registered
    def __new__(cls, name, vname=None, xf=None, rel=None, bins=[],
                bins_labels=None, weight=1, at_least=1, inj=False):
        if name in coverage_db:
            return coverage_db[name]
        else:
            return super(CoverPoint, cls).__new__(CoverPoint)

    def __init__(self, name, vname=None, xf=None, rel=None, bins=[],
                 bins_labels=None, weight=1, at_least=1, inj=False):
        if not name in coverage_db:
            CoverItem.__init__(self, name)
            if self._parent is None:
                raise Exception("CoverPoint must have a parent \
                                 (parent.CoverPoint)")

            if (bins_labels is not None) and (len(bins_labels) != len(bins)):
                raise Exception("Length of bins and bins_labels must be \
                                 equal")

            self._bins_labels = bins_labels

            self._transformation = xf
            self._vname = vname

            # equality operator is the default bins matching relation
            self._relation = rel if rel is not None else operator.eq
            self._weight = weight
            self._at_least = at_least
            self._injection = inj

            if (len(bins) != 0):
                self._size = self._weight * len(bins)
                self._hits = OrderedDict.fromkeys(bins, 0)
            else:  # if no bins specified, add one bin equal True
                self._size = self._weight
                self._hits = OrderedDict.fromkeys([True], 0)

            #make a map assigning label to the bin
            if self._bins_labels is not None:
                self._labels_bins = dict(zip(bins, bins_labels))

            # determines whether decorated a bound method
            self._decorates_method = None
            # determines whether transformation function is a bound method
            self._trans_is_method = None
            self._parent._update_size(self._size)

            self._new_hits = []  # list of bins hit per single function call

    def __call__(self, f):
        @wraps(f)
        def _wrapped_function(*cb_args, **cb_kwargs):

            # if transformation function not defined, simply return arguments
            if self._transformation is None:
                if self._vname is None:
                    def dummy_f(*cb_args):  # return a tuple or single object
                        if len(cb_args) > 1:
                            return cb_args
                        else:
                            return cb_args[0]
                # if vname defined, match it to the decroated function args
                else:
                    arg_names = list(inspect.signature(f).parameters)
                    idx = arg_names.index(self._vname)

                    def dummy_f(*cb_args):
                        return cb_args[idx]

                self._transformation = dummy_f

            # for the first time only check if decorates method in the class
            if self._decorates_method is None:
                self._decorates_method = False
                for x in inspect.getmembers(cb_args[0]):
                    if '__func__' in dir(x[1]):
                        # compare decorated function name with class functions
                        self._decorates_method = \
                            f.__name__ == x[1].__func__.__name__
                        if self._decorates_method:
                            break

            # for the first time only check if a transformation function is a
            # method
            if self._trans_is_method is None:
                self._trans_is_method = "self" in inspect.signature(
                    self._transformation).parameters

            current_coverage = self.coverage
            self._new_hits = []

            # if function is bound then remove "self" from the arguments list
            if self._decorates_method ^ self._trans_is_method:
                result = self._transformation(*cb_args[1:])
            else:
                result = self._transformation(*cb_args)

            # compare function result using relation function with matching
            # bins
            for bin in self._hits:
                if self._relation(result, bin):
                    self._hits[bin] += 1
                    if self._bins_labels is not None:
                        self._new_hits.append(self._labels_bins[bin])
                    else:
                        self._new_hits.append(bin)
                    # check bins callbacks
                    if bin in self._bins_callbacks:
                        self._bins_callbacks[bin]()
                    # if injective function, continue through all bins
                    if not self._injection:
                        break

            # notify parent about new coverage level
            self._parent._update_coverage(self.coverage - current_coverage)

            # check threshold callbacks
            for ii in self._threshold_callbacks:
                if (ii > 100 * current_coverage / self.size
                        and ii <= 100 * self.coverage / self.size):
                    self._threshold_callbacks[ii]()

            return f(*cb_args, **cb_kwargs)
        return _wrapped_function

    @property
    def coverage(self):
        coverage = self._size
        for ii in self._hits:
            if self._hits[ii] < self._at_least:
                coverage -= self._weight
        return coverage

    @property
    def detailed_coverage(self):
        if self._bins_labels is not None:
            return dict(zip(self._bins_labels, list(self._hits.values())))
        return self._hits


class CoverCross(CoverItem):
    """Class used to create coverage crosses as decorators.

    This decorator samples members of the decorated function (its signature). 
    It matches tuples cross-bins which are Cartesian products of bins defined 
    in :class:`CoverPoints <CoverPoint>` (items).

    Args:
        name (str): a ``CoverCross`` path and name, defining its position in a 
            coverage trie.
        items (list): a list of :class:`CoverPoints <CoverPoint>` by names, 
            to create a Cartesian product of cross-bins.
        ign_bins (list, optional): a list of bins to be ignored.
        weight (int, optional): a ``CoverCross`` weight (by default ``1``).
        at_least (int, optional): the number of hits per bin to be considered 
            as covered (by default ``1``).

    Example:

    >>> @coverage.CoverPoint(
    ...     name = "top.parent.coverpoint1", 
    ...     xf = lambda x, y: x, 
    ...     bins = range(1, 5) # 5 bins in total
    ... )
    >>> @coverage.CoverPoint(
    ...     name = "top.parent.coverpoint2",
    ...     xf = lambda x, y: y, 
    ...     bins = range(1, 5) # 5 bins in total
    ... )
    >>> @coverage.CoverCross(
    ...     name = "top.parent.covercross", 
    ...     items = ["top.parent.coverpoint1", "top.parent.coverpoint2"],
    ...     ign_bins = [(1, 1), (5, 5)], # 5x5 - 2 = 23 bins in total
    ... )
    >>> def decorated_func(self, arg_a, arg_b):
    >>> # bin from the bins list [(1, 2), (1, 3)...(5, 4)] will be matched 
    >>> # when a tuple (x=arg_a, y=arg_b) was sampled at this function call.
    ...     ...
    """

    # conditional Object creation, only if name not already registered
    def __new__(cls, name, items=[], ign_bins=[], weight=1, at_least=1):
        if name in coverage_db:
            return coverage_db[name]
        else:
            return super(CoverCross, cls).__new__(CoverCross)

    def __init__(self, name, items=[], ign_bins=[], weight=1, at_least=1):
        if not name in coverage_db:
            CoverItem.__init__(self, name)
            if self._parent is None:
                raise Exception("CoverCross must have a parent \
                                 (parent.CoverCross)")

            self._weight = weight
            self._at_least = at_least
            # equality operator is the defult ignore bins matching relation
            self._items = items

            bins_lists = []
            for cp_names in self._items:
                bins_lists.append(
                    coverage_db[cp_names].detailed_coverage.keys())

            # a map of cross-bins, key is a tuple of bins Cartesian product
            self._hits = dict.fromkeys(itertools.product(*bins_lists), 0)

            # remove ignore bins from _hits map if relation is true
            for x_bin in list(self._hits.keys()):
                for ignore_bins in ign_bins:
                    remove = True
                    for ii in range(0, len(x_bin)):
                        if ignore_bins[ii] is not None:
                            if (ignore_bins[ii] != x_bin[ii]):
                                remove = False
                    if remove and (x_bin in self._hits):
                        del self._hits[x_bin]

            self._size = self._weight * len(self._hits)
            self._parent._update_size(self._size)

    def __call__(self, f):
        @wraps(f)
        def _wrapped_function(*cb_args, **cb_kwargs):

            current_coverage = self.coverage

            hit_lists = []
            for cp_name in self._items:
                hit_lists.append(coverage_db[cp_name]._new_hits)

            # a list of hit cross-bins, key is a tuple of bins Cartesian
            # product
            for x_bin_hit in list(itertools.product(*hit_lists)):
                if x_bin_hit in self._hits:
                    self._hits[x_bin_hit] += 1
                    # check bins callbacks
                    if x_bin_hit in self._bins_callbacks:
                        self._bins_callbacks[x_bin_hit]()

            # notify parent about new coverage level
            self._parent._update_coverage(self.coverage - current_coverage)

            # check threshold callbacks
            for ii in self._threshold_callbacks:
                if (ii > 100 * current_coverage / self.size
                        and ii <= 100 * self.coverage / self.size):
                    self._threshold_callbacks[ii]()

            return f(*cb_args, **cb_kwargs)
        return _wrapped_function

    @property
    def coverage(self):
        coverage = self._size
        for ii in self._hits:
            if self._hits[ii] < self._at_least:
                coverage -= self._weight
        return coverage

    @property
    def detailed_coverage(self):
        return self._hits


class CoverCheck(CoverItem):
    """Class used to create coverage checks as decorators. 

    It is a simplified :class:`CoverPoint` with defined 2 bins:
    *PASS* and *FAIL* and ``f_pass()`` and ``f_fail()`` functions. 

    Args:
        name (str): a ``CoverCheck`` path and name, defining its position in a 
            coverage trie.
        f_fail: a failure condition function - if it returns ``True``, them
            coverage level is set to ``0`` permanently.
        f_pass: a pass condition function - if it returns ``True``, the 
            coverage level is set to ``weight`` after ``at_least`` hits. 
        weight (int, optional): a ``CoverCheck`` weight (by default ``1``).
        at_least (int, optional): the number of hits of the ``f_pass`` function 
            to consider a particular ``CoverCheck`` as covered. 

    Example:

    >>> @coverage.CoverCheck(
    ...     name = "top.parent.check", 
    ...     f_fail = lambda x : x == 0, 
    ...     f_pass = lambda x : x < 5)
    >>> def decorated_fun(self, arg):
    >>> # CoverCheck is 100% covered when (arg < 5) and never (arg == 0) was 
    >>> # sampled. CoverCheck is set to 0 unconditionally when at least once
    >>> # (arg == 0) was sampled.
    ...     ...
    """

    # conditional Object creation, only if name not already registered
    def __new__(cls, name, f_fail, f_pass=None, weight=1, at_least=1):
        if name in coverage_db:
            return coverage_db[name]
        else:
            return super(CoverCheck, cls).__new__(CoverCheck)

    def __init__(self, name, f_fail, f_pass=None, weight=1, at_least=1):
        if not name in coverage_db:
            CoverItem.__init__(self, name)
            if self._parent is None:
                raise Exception("CoverCheck must have a parent \
                                 (parent.CoverCheck)")
            self._weight = weight
            self._at_least = at_least
            self._f_pass = f_pass
            self._f_fail = f_fail
            self._size = weight
            self._hits = dict.fromkeys(["PASS", "FAIL"], 0)

            # determines whether decorated a bound method
            self._decorates_method = None
            # determines whether pass function is a bound method
            self._f_pass_is_method = None
            # determines whether fail function is a bound method
            self._f_fail_is_method = None
            self._parent._update_size(self._size)

    def __call__(self, f):
        @wraps(f)
        def _wrapped_function(*cb_args, **cb_kwargs):

            # if pass function not defined always return True
            if self._f_pass is None:
                def dummy_f(*cb_args):
                    return True
                self._f_pass = dummy_f

            # for the first time only check if decorates method in the class
            if self._decorates_method is None:
                self._decorates_method = False
                for x in inspect.getmembers(cb_args[0]):
                    if '__func__' in dir(x[1]):
                        # compare decorated function name with class functions
                        self._decorates_method = f.__name__ == x[
                            1].__func__.__name__
                        if self._decorates_method:
                            break

            # for the first time only check if a pass/fail function is a method
            if self._f_pass_is_method is None and self._f_pass:
                self._f_pass_is_method = "self" in inspect.signature(
                    self._f_pass).parameters
            if self._f_fail_is_method is None:
                self._f_fail_is_method = "self" in inspect.signature(
                    self._f_fail).parameters

            current_coverage = self.coverage

            # may be False (failed), True (passed) or None (undetermined)
            passed = None

            # if function is bound then remove "self" from the arguments list
            if self._decorates_method ^ self._f_pass_is_method:
                passed = True if self._f_pass(*cb_args[1:]) else None
            else:
                passed = True if self._f_pass(*cb_args) else None

            if self._decorates_method ^ self._f_fail_is_method:
                passed = False if self._f_fail(*cb_args[1:]) else passed
            else:
                passed = False if self._f_fail(*cb_args) else passed

            if passed:
                self._hits["PASS"] += 1
            elif passed is not None:
                self._hits["FAIL"] += 1

            if passed is not None:

                # notify parent about new coverage level
                self._parent._update_coverage(self.coverage - current_coverage)

                # check threshold callbacks
                for ii in self._threshold_callbacks:
                    if (ii > 100 * current_coverage / self.size
                            and ii <= 100 * self.coverage / self.size):
                        self._threshold_callbacks[ii]()

                # check bins callbacks
                if "PASS" in self._bins_callbacks and passed:
                    self._bins_callbacks["PASS"]()
                elif "FAIL" in self._bins_callbacks and not passed:
                    self._bins_callbacks["FAIL"]()

            return f(*cb_args, **cb_kwargs)
        return _wrapped_function

    @property
    def coverage(self):
        coverage = 0
        if self._hits["FAIL"] == 0 and self._hits["PASS"] >= self._at_least:
            coverage = self._weight
        return coverage

    @property
    def detailed_coverage(self):
        return self._hits


def coverage_section(*coverItems):
    """Combine multiple coverage items into a single decorator.

    Args:
        *coverItems ((multiple) :class:`CoverItem`): coverage primitives to be
            combined.

    Example:

    >>> my_coverage = coverage.coverageSection(
    ...     coverage.CoverPoint("x", ...),
    ...     coverage.CoverPoint("y", ...),
    ...     coverage.CoverCross("z", ...),
    ...     ...
    ... )
    >>>
    >>> @my_coverage
    >>> def decorated_fun(self, arg):
    ...     ...
    """
    def _nested(*decorators):
        def _decorator(f):
            for dec in reversed(*decorators):
                f = dec(f)
            return f
        return _decorator

    return _nested(coverItems)

### XML coverage database functions

# XML pretty print format - ElementTree lib extension
def _indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            _indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

def XML_merger(merged_xml_name, *xmls):
    """ Function used for merging coverage metrics in XML format.

    Args:
        merged_xml_name (str): output XML name with .xml suffix
        *xmls ((multiple) str): comma separated XML names with .xml suffix

    Example:

    >>> XML_merger('merged.xml', 'one.xml', 'other.xml') # merge one and other
    """
    roots = [et.parse(xml).getroot() for xml in xmls]
    merged_root = copy.deepcopy(roots[0])
    # abs name to element mapping - merged items
    name_to_elem = {el.attrib['abs_name']: el for el in merged_root.iter()}

    def combine():
        for root in roots[1:]:
            combine_element(root)
        _indent(merged_root) # pretty print format
        et.ElementTree(merged_root).write(merged_xml_name)

    def combine_element(root):
        # Elements not present in merged items
        new_elements = [elem for elem in root.iter()
                        if elem.attrib['abs_name'] not in name_to_elem.keys()]
        # Sort descending
        new_elements.sort(key=lambda _: _.attrib['abs_name'].count('.'))
        # Bin list to be updated w/o bins not present in the merged
        bin_list = [elem for elem in root.iter() if 'bin' in elem.tag
                    and elem not in new_elements]

        def get_parent(abs_name):
            return '.'.join(abs_name.split('.')[:-1])

        def update_parent(name, bin_update=False, new_element_update=False,
                          coverage_update=0, size_update=0):
            parent_name = get_parent(name)
            if parent_name == '':
                return # Top reached
            else:
                if new_element_update:
                    coverage_update = int(
                        name_to_elem[name].attrib['coverage'])
                    size_update = int(
                        name_to_elem[name].attrib['size'])
                elif bin_update:
                    coverage_update = int(
                        name_to_elem[parent_name].attrib['weight'])

                # Update current parent
                name_to_elem[parent_name].attrib['coverage'] = str(int(
                    name_to_elem[parent_name].attrib['coverage'])
                    + coverage_update)
                name_to_elem[parent_name].attrib['size'] = str(int(
                    name_to_elem[parent_name].attrib['size'])+size_update)
                name_to_elem[parent_name].attrib['cover_percentage'] = str(
                    round((int(name_to_elem[parent_name].attrib['coverage'])
                    *100/int(name_to_elem[parent_name].attrib['size'])), 2))
                # Recursively update parents
                update_parent(parent_name, False, False,
                              coverage_update, size_update)

        # Merge function body
        # Extend merged with new elements in descending order
        for elem in new_elements:
            abs_name = elem.attrib['abs_name']
            parent_name = get_parent(abs_name)
            name_to_elem[abs_name] = et.SubElement(
                name_to_elem[parent_name], elem.tag, attrib=elem.attrib)
            if 'bin' not in elem.tag:
                update_parent(name=abs_name, bin_update=False, 
                              new_element_update=True)

        # Update bins already present in the XML
        for bin_element in bin_list:
            hits = int(bin_element.attrib['hits'])
            if hits > 0:
                abs_name = bin_element.attrib['abs_name']
                hits_orig = int(name_to_elem[abs_name].attrib['hits'])
                # Update the bin value
                name_to_elem[abs_name].attrib['hits'] = str(hits+hits_orig)
                # Check if upstream needs updating
                parent_name = get_parent(abs_name)
                parent_hits_threshold = int(
                    name_to_elem[parent_name].attrib['at_least'])
                if (hits_orig < parent_hits_threshold
                        and hits_orig+hits >= parent_hits_threshold):
                    update_parent(name=abs_name, bin_update=True,
                                  new_element_update=False)
    
    # Call combine function to merge XMLs
    combine()

# deprecated

def reportCoverage(logger, bins=False):
    """.. deprecated:: 1.0"""
    warnings.warn(
        "Function reportCoverage() is deprecated, use "
        + "coverage_db.report_coverage() instead"
    )
    coverage_db.report_coverage(logger, bins)


def coverageSection(*coverItems):
    """.. deprecated:: 1.0"""
    warnings.warn(
        "Function coverageSection() is deprecated, use coverage_section() instead"
    )
    return coverage_section(*coverItems)


